import json, secretstorage

from datetime import datetime
from functools import wraps
from secretstorage.collection import Collection, get_collection_by_alias, create_collection
from secretstorage.exceptions import ItemNotFoundException, LockedException
from secretstorage.item import Item

from .base import PasswordItem, PasswordStorage, PasswordResult

from ..functions import TimeoutCallback, run_in_gui_thread


def lock_check(func):
	@wraps(func)
	def wrapper(self, *args, **kwargs):
		if not self.active:
			raise ConnectionError('No active connection to Gnome Keyring')

		if self.locked:
			self.unlock()

		return func(self, *args, **kwargs)
	return wrapper


class GnomeKeyringItem(PasswordItem):
	def __init__(self, *args, **kwargs):
		PasswordItem.__init__(self, *args, **kwargs)
		self._attrs = self._data.get_attributes()


	def _get_value(self, key):
		if key == 'id':
			return int(Path(self._data.item_path).name)

		elif key == 'label':
			if not (label := self._data.get_label()):
				return f'{self.username} @ {self.domain}'

			return label

		elif key == 'password':
			return self._data.get_secret().decode()

		elif key == 'created':
			return datetime.fromtimestamp(self._data.get_created())

		elif key == 'modified':
			return datetime.fromtimestamp(self._data.get_modified())

		return self._attrs.get(key)


	def _set_value(self, key, value):
		if key == 'label':
			self._data.set_label(value)
			return

		elif key == 'password':
			self._data.set_secret(value.encode())
			return

		elif key in {'created', 'modified'}:
			return

		self._attrs[key] = value
		self._data.set_attributes(self._attrs)


	def delete(self):
		self._data.delete()


	def update(self, _data={}, **kwargs):
		for key, value in [*_data.items(), *kwargs.items()]:
			if key not in self.keys():
				raise KeyError(f'Invalid PasswordItem key: {key}')

			self[key] = value


class GnomeKeyringStorage(PasswordStorage):
	#result_class = PasswordResult
	item_class = GnomeKeyringItem


	def __init__(self, profile):
		self.profile = profile
		self.connection = None
		self.collection = None


	@property
	def active(self):
		return True if self.connection else False


	@property
	def locked(self):
		if not self.active:
			return True

		return self.collection.is_locked()


	@property
	def name(self):
		if self.profile == 'BarksharkWeb':
			return self.profile

		else:
			return f'BarksharkWeb_{self.profile}'


	@property
	def path(self):
		return Path('/org/freedesktop/secrets/collection').join(self.name)


	def start(self):
		if self.connection and self.collection:
			return

		if not self.connection:
			self.connection = secretstorage.dbus_init()

		if not secretstorage.check_service_availability(self.connection):
			self.disconnect()
			raise ConnectionError('Failed to connect to Secret Service server')

		return self.get_keyring()


	def stop(self):
		if self.connection:
			self.connection.close()

		self.connection = None
		self.collection = None


	def get_keyring(self):
		if not self.collection:
			try:
				self.collection = Collection(self.connection, self.path)
			except ItemNotFoundException:
				create_collection(self.connection, self.name)
				self.collection = Collection(self.connection, self.path)

		return self.collection


	def unlock(self):
		return self.collection.unlock()


	def lock(self):
		return self.collection.lock()


	def create_row(self, *args, **kwargs):
		return self.item_class.new(*args, **kwargs, storage=self)


	@lock_check
	def fetch(self, id=None, label=None, **kwargs):
		items = []

		if id:
			try:
				item = Item(self.connection, self.path.join(str(id)), self.collection.session)
				return GnomeKeyringItem(self, item)

			except ItemNotFoundException:
				raise KeyError(f'Cannot find password with id: {id}') from None

		elif label:
			for item in self.collection.get_all_items():
				if label.lower() in item.get_label().lower():
					return GnomeKeyringItem(self, item)

			raise KeyError(f'Cannot find password with text in label: {label}')

		elif not len(kwargs):
			items = self.collection.get_all_items()

		else:
			items = self.collection.search_items(kwargs)

		return self.result_class(self, items)


	@lock_check
	def insert(self, username, password, url, label=None, **kwargs):
		url = Url(url)

		data = {
			'label': label or f'{username} @ {url.domain}',
			'username': username,
			'domain': url.domain,
			'url': url
		}

		return self.collection.create_item(label, data, password.encode())


	@lock_check
	def update(self, row_id, **kwargs):
		item = self.fetch(row_id)
		item.update(**kwargs)

		return item


	@lock_check
	def update_row(self, item, **kwargs):
		item.update(**kwargs)

		return item


	@lock_check
	def remove(self, row_id):
		item = self.fetch(row_id)
		item.delete()

		return True


	@lock_check
	def remove_row(self, item):
		item.delete()

		return True

